from typing import Optional

import marshmallow_dataclass
from marshmallow import pre_load

from datagen.modalities.textual.common.ndarray import NumpyArray


@marshmallow_dataclass.dataclass
class Orientation:
    look_at_vector: NumpyArray
    up_vector: NumpyArray


@marshmallow_dataclass.dataclass
class FOV:
    horizontal: float
    vertical: float


@marshmallow_dataclass.dataclass
class Sensor:
    width: float
    height: float

    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs):
        in_data["width"] = in_data.pop("sensor_width")
        in_data["height"] = in_data.pop("sensor_height")
        return in_data


@marshmallow_dataclass.dataclass
class CameraMetadata:
    camera_name: str
    camera_type: str
    focal_length: Optional[float]
    sensor: Sensor
    resolution_px: NumpyArray
    aspect_px: NumpyArray
    location: NumpyArray
    orientation: Orientation
    fov: FOV
    intrinsic_matrix: NumpyArray
    extrinsic_matrix: NumpyArray



from typing import TypeVar, Union

import numpy as np

from datagen.api.assets import Point, Vector, Rotation

CA = TypeVar("CA", bound=Union[Point, Vector, Rotation, np.ndarray])


def normalize(vector: np.ndarray) -> np.ndarray:
    return vector / np.linalg.norm(vector)


def to_arrays(*computationals: CA):
    ndarrays = []
    for c in computationals:
        if isinstance(c, ComputationalAsset):
            ndarrays.append(c.to_ndarray())
        elif isinstance(c, np.ndarray):
            ndarrays.append(c)
        else:
            raise ValueError()
    return ndarrays



import inspect
import sys


class MutuallyExclusiveArgsError(ValueError):
    ...


def mutually_exclusive(*mutually_exc_args_names: str):
    def wrapper(f):
        def inner(*args, **kwargs):
            bound_args = inspect.signature(f).bind(*args, **kwargs)
            bound_args.apply_defaults()
            validate(bound_args.arguments)
            return f(*args, **kwargs)

        def validate(kwargs: dict) -> None:
            mutually_exclusive_args_count = get_mutually_exc_args_count(kwargs)
            if mutually_exclusive_args_count != 1:
                raise MutuallyExclusiveArgsError("Mutually exclusive arguments must have exactly one value provided!")

        def get_mutually_exc_args_count(kwargs):
            mutually_exc_args_count = 0
            for arg_name, arg_val in kwargs.items():
                if arg_name in mutually_exc_args_names and arg_val is not None:
                    mutually_exc_args_count += 1
            return mutually_exc_args_count

        return inner

    return wrapper


sys.modules[__name__] = mutually_exclusive  # type: ignore



from typing import List, Tuple, Any
import sys

if sys.version_info < (3, 10):
    from importlib_metadata import entry_points, EntryPoint  # type: ignore
else:
    from importlib.metadata import entry_points, EntryPoint  # type: ignore

from dependency_injector import providers


NOT_BASE_CLASS_PLUGIN = "<NOT_BASE_CLASS_PLUGIN>"


class PluginsFactory(providers.Provider):

    __slots__ = ("_factory", "_base_class")

    def __init__(self, base_class, *args, **kwargs):
        self._base_class = base_class
        self._factory = providers.Factory(self._get_provided_type(), *args, **kwargs)
        super().__init__()

    def _get_provided_type(self) -> type:
        plugins_classes = self._collect_plugins_classes()
        return self._create_type(plugins_classes) if len(plugins_classes) > 0 else self._base_class

    def _collect_plugins_classes(self) -> Tuple[type]:
        plugins_classes = []
        for group_name, entrypoints in entry_points().items():
            if self._is_base_class_plugins_group(group_name):
                plugins_classes.extend(self._load_plugins(plugins_entrypoints=entrypoints))
        return tuple(plugins_classes)

    def _is_base_class_plugins_group(self, plugins_group_name: str) -> bool:
        return self._base_class.PLUGINS_GROUP_NAME.startswith(plugins_group_name)

    def _load_plugins(self, plugins_entrypoints: List[EntryPoint]) -> List[type]:
        plugin_classes = []
        for entrypoint in plugins_entrypoints:
            plugin = entrypoint.load()
            if self._is_base_class_plugin(plugin):
                plugin_classes.append(plugin)
        return plugin_classes

    def _create_type(self, plugins: Tuple[type]) -> type:
        name, dict_ = self._base_class.__name__, {"PLUGINS_GROUP_NAME": self._base_class.PLUGINS_GROUP_NAME}
        if self._base_class_specified_in(plugins):
            return type(name, plugins, dict_)
        else:
            # In case all dev extend some parent of the base class but not the base class itself,
            # The base class must be provided explicitly.
            return type(name, (self._base_class, *plugins), dict_)

    def _base_class_specified_in(self, plugins: Tuple[type]) -> bool:
        return any(issubclass(p, self._base_class) for p in plugins)

    def __deepcopy__(self, memo):
        copied = memo.get(id(self))
        if copied is not None:
            return copied

        copied = self.__class__(
            self._factory.provides,
            *providers.deepcopy(self._factory.args, memo),
            **providers.deepcopy(self._factory.kwargs, memo),
        )
        self._copy_overridings(copied, memo)

        return copied

    @property
    def related(self):
        """Return related providers generator."""
        yield from [self._factory]
        yield from super().related

    def _provide(self, args, kwargs):
        return self._factory(*args, **kwargs)

    def _is_base_class_plugin(self, plugin: Any) -> bool:
        plugin_group_name = getattr(plugin, "PLUGINS_GROUP_NAME", NOT_BASE_CLASS_PLUGIN)
        return self._base_class.PLUGINS_GROUP_NAME.startswith(plugin_group_name)



import asyncio
import builtins
from typing import Any

from datagen.core.tasks.task import Task


class TaskRunner:
    @classmethod
    def run(cls, task: Task, data: Any) -> Any:
        task.setup()

        if hasattr(builtins, "__IPYTHON__"):
            # In case of running in a IPYTHON environment
            import nest_asyncio

            nest_asyncio.apply()

        response = asyncio.run(cls._internal_run(task=task, data=data))
        return response

    @classmethod
    async def _internal_run(cls, task: Task, data: Any) -> Any:
        await task.input_channel.put(data)
        await task.execute()
        return await task.output_channel.take()



import json
from functools import lru_cache, partial

from dependency_injector import containers, providers
from packaging import version

INITIAL_MODALITIES_VERSION = "1.0.0"


@lru_cache(maxsize=None)
def modality_factory(modality_container: providers.Container, modality_file_path: str):
    """
    Important - Please note this function caches its return values!
    """
    modality_dict, modality_version = _parse_modality_file(modality_file_path)
    return modality_container.create(modality_version, modality_dict=modality_dict)


def _parse_modality_file(modality_file_path: str) -> tuple:
    with open(modality_file_path) as f:
        modality_params = json.load(f)
        modality_version = _get_modality_version(modality_params)
    return modality_params, modality_version


def _get_modality_version(modality_params: dict) -> int:
    modality_version = modality_params.pop("version", INITIAL_MODALITIES_VERSION)
    modality_version_ = version.parse(modality_version).major
    return modality_version_


def load_dataclass(clazz: type, modality_dict: dict):
    return clazz.Schema().load(modality_dict)


modality_dataclass_factory = partial(providers.Factory, load_dataclass)


def load_pydantic_obj(clazz: type, modality_dict: dict):
    return clazz.parse_obj(modality_dict)


modality_pydantic_factory = partial(providers.Factory, load_pydantic_obj)


class CameraMetadataModalityContainer(containers.DeclarativeContainer):

    from .camera_metadata import CameraMetadata

    create = providers.FactoryAggregate({1: modality_dataclass_factory(clazz=CameraMetadata)})


class SegmentationModalityContainer(containers.DeclarativeContainer):

    from .segmentation import Segmentation

    create = providers.FactoryAggregate({1: modality_dataclass_factory(clazz=Segmentation)})


class EnvironmentsModalityContainer(containers.DeclarativeContainer):

    from .environments import Environments

    create = providers.FactoryAggregate({1: modality_dataclass_factory(clazz=Environments)})


class LightsMetadataModalityContainer(containers.DeclarativeContainer):

    from datagen.modalities.textual.base.lights_metadata import LightsMetadata

    create = providers.FactoryAggregate({1: modality_dataclass_factory(clazz=LightsMetadata)})


class BaseModalitiesContainer(containers.DeclarativeContainer):
    environments = providers.Callable(
        modality_factory, modality_container=providers.Container(EnvironmentsModalityContainer)
    )

    camera_metadata = providers.Callable(
        modality_factory, modality_container=providers.Container(CameraMetadataModalityContainer)
    )

    segmentation = providers.Callable(
        modality_factory, modality_container=providers.Container(SegmentationModalityContainer)
    )

    lights_metadata = providers.Callable(
        modality_factory, modality_container=providers.Container(LightsMetadataModalityContainer)
    )



import json
from pathlib import Path
from typing import List, Optional, Union

from datagen.api.assets import Background, Camera, DataRequest, Glasses, Human, HumanDatapoint, Light, Mask
from datagen.api.client.schemas import DataResponse, DataResponseStatus, DownloadRequest, DownloadURL
from datagen.api.requests.datapoint.builder import HumanDatapointBuilder
from datagen.api.requests.director import DataRequestDirector
from datagen.config import settings
from datagen.core.tasks import TaskContainer
from datagen.core.tasks.task_runner import TaskRunner
from datagen.dev.logging import get_logger

logger = get_logger(__name__)

DEFAULT_DATAPOINT_REQUESTS_JSON_NAME = "datagen_data_request.json"


class DatagenAPI:
    def __init__(self):
        self._request_director = DataRequestDirector()
        self._task_container = TaskContainer()

    def create_datapoint(
        self,
        human: Human,
        camera: Camera,
        glasses: Optional[Glasses] = None,
        mask: Optional[Mask] = None,
        background: Optional[Background] = None,
        lights: Optional[List[Light]] = None,
    ) -> HumanDatapoint:
        self._request_director.builder = HumanDatapointBuilder(human, camera, glasses, mask, background, lights)
        return self._request_director.build_datapoint()

    def load(self, path: Union[Path, str]) -> DataRequest:
        if not isinstance(path, Path):
            path = self._get_request_json_path(path)
        path = self._get_request_json_path(path)
        return DataRequest.parse_file(path)

    def generate(
        self,
        request: DataRequest,
        generation_name: str,
    ) -> DataResponse:
        batches = DatagenAPI.batch_request(request=request)
        return TaskRunner().run(
            task=self._task_container.pipeline_factory.data_generation(number_of_batches=len(batches)),
            data=(batches, generation_name),
        )

    def stop(self, generation_id) -> None:
        TaskRunner().run(
            task=self._task_container.client_task(task_name="stop_generation"),
            data=generation_id,
        )

    def get_status(self, generation_id: str) -> DataResponseStatus:
        return TaskRunner().run(
            task=self._task_container.client_task(task_name="status"),
            data=generation_id,
        )

    def get_download_urls(self, generation_id: str) -> List[DownloadURL]:
        return TaskRunner().run(
            task=self._task_container.client_task(task_name="get_download_links"),
            data=generation_id,
        )

    def download(
        self, urls: List[DownloadURL], dest_folder: str, dataset_name: str, remove_tar_files: bool = True
    ) -> None:
        TaskRunner().run(
            task=self._task_container.pipeline_factory.download(
                number_of_files=len(urls), remove_tar_files=remove_tar_files
            ),
            data=DownloadRequest(urls=urls, path=dest_folder, dataset_name=dataset_name).batch(),
        )

    def dump(self, request: DataRequest, path: Union[Path, str] = None) -> None:
        if not isinstance(path, Path):
            path = self._get_request_json_path(path)
        path.write_text(json.dumps(request.dict(), indent=3, sort_keys=True))
        logger.info(
            f"Request was successfully dumped to path '{path.absolute()}' ({len(request)} datapoints total).",
        )

    def _get_request_json_path(self, path: Optional[str]) -> Path:
        path_ = Path(path) if path else Path.cwd().joinpath(DEFAULT_DATAPOINT_REQUESTS_JSON_NAME)
        if not path_.parent.exists():
            raise FileNotFoundError(f"{path_.parent} folder does not exist, cannot dump requests")
        path_.touch()
        return path_

    @staticmethod
    def batch_request(request: DataRequest) -> List[DataRequest]:
        batch_size = settings["batch_size"]
        batches = []
        num_of_datapoints = len(request.datapoints)
        for dp_idx in range(0, num_of_datapoints, batch_size):
            batches.append(DataRequest(datapoints=request.datapoints[dp_idx : dp_idx + batch_size]))

        return batches



from dependency_injector import containers, providers

from datagen.api.impl import DatagenAPI
from datagen.dev.logging import get_logger

logger = get_logger(__name__)


class ApiContainer(containers.DeclarativeContainer):

    api = providers.Singleton(DatagenAPI)



from typing import Dict, List

from datagen_protocol.schema import DataSequence
from pydantic.config import Extra
from pydantic.main import BaseModel

from datagen.api.assets import ExtrinsicParams
from datagen.api.catalog.impl import InitParametersHook


class HICPresetsHook(InitParametersHook[DataSequence]):
    def __init__(self, asset_id_to_asset_presets: dict):
        self._asset_id_to_asset_presets = asset_id_to_asset_presets

    def __call__(self, asset_id: str) -> dict:
        return dict(presets=self._asset_id_to_asset_presets[asset_id])




from dependency_injector import containers, providers

from datagen.dev import PluginsFactory
from datagen.components.datapoint import DatapointsRepository
from datagen.components.datapoint import HICDataPoint, IdentitiesDataPoint
from datagen.modalities.containers import DatapointModalitiesContainer


class DatapointsContainer(containers.DeclarativeContainer):

    __self__ = providers.Self()

    config = providers.Configuration()

    modalities = providers.Container(DatapointModalitiesContainer, config=config)

    factory = providers.Selector(
        config.environment,
        hic=PluginsFactory(HICDataPoint, modalities_container=modalities),
        identities=PluginsFactory(IdentitiesDataPoint, modalities_container=modalities),
    )

    repo = providers.Factory(DatapointsRepository, __self__)



import logging
import sys

FORMATTER = logging.Formatter("%(asctime)s - %(name)s - %(levelname)s - %(message)s")


def get_console_handler():
    console_handler = logging.StreamHandler(sys.stdout)
    console_handler.setFormatter(FORMATTER)
    return console_handler


def get_logger(logger_name: str):
    logger = logging.getLogger(name=logger_name)
    logger.setLevel(level=logging.INFO)
    logger.addHandler(hdlr=get_console_handler())
    logger.propagate = False
    return logger



import enum
import sys
from pathlib import Path
from typing import List, Optional, Union
from urllib.parse import urlsplit

from pydantic import BaseModel, Field, validator

from datagen.config import settings


class DataResponse(BaseModel):
    generation_name: str = Field(
        description="Generation name provided by Datagen platform. In a format " "of 'Field - <date>."
    )
    generation_id: str = Field(
        description="Unique UUID provided by Datagen platform. With the generation id you'll "
        "be able to query generation_status, stop generation and get download urls"
    )
    dgu_hour: float = Field(description="The DGU Hour cost of the requested generation.")
    renders: int = Field(description="The number of datapoints received in this request.")
    scenes: int = Field(description="The number of scenes received in this request.")


class EGenerationStatus(enum.Enum):
    START = "START"
    IN_PROGRESS = "IN-PROGRESS"
    PENDING = "IN-PENDING"
    COMPLETED = "COMPLETED"
    FAILED = "FAILED"


class DownloadURL(BaseModel):
    url: str
    filename: str
    dataset_name: Optional[str] = None

    @validator("url")
    def validate_url_host(cls, val: str) -> str:
        hostname = urlsplit(val).hostname
        if hostname in settings["url"]["download"]["hostnames"]:
            raise ValueError("Received invalid hostname.")
        return val

    @classmethod
    def parse_url(cls, url: str, dataset_name: Optional[str] = None) -> "DownloadURL":
        path = urlsplit(url).path.split("/")
        filename = path[-1]
        return cls(url=url, filename=filename, dataset_name=dataset_name)


class DownloadRequest(BaseModel):
    urls: List[DownloadURL]
    path: str
    dataset_name: str

    @validator("path")
    def validate_url_host(cls, path: str) -> str:
        if not Path(path).exists():
            Path(path).mkdir(parents=True)
        return path

    def batch(self) -> list:
        batches = []
        for url in self.urls:
            url.filename = str(Path(self.path, url.filename))
            url.dataset_name = self.dataset_name
            batches.append(url)

        return batches


class DataResponseStatus(BaseModel):
    status: EGenerationStatus = Field(description="Current status of the generation process.")
    percentage: Optional[int] = Field(description="The percentage of data generation completion.", ge=0, le=100)
    estimation_time_ms: Optional[Union[int, str]] = Field(
        description="Estimated time in ms to data generation completion.", ge=0
    )

    @validator("estimation_time_ms")
    def validate_name(cls, val: Union[int, str]):
        if not val:
            return sys.maxsize
        return val


class ErrorDetail(BaseModel):
    loc: list
    msg: str
    type: str


class ValidationError(BaseModel):
    error_name: str
    description: str
    text: str
    details: List[ErrorDetail]


class ErrorResponse(BaseModel):
    error: Union[str, ValidationError]

    def __str__(self):
        if isinstance(self.error, str):
            return f"Error: {self.error}"
        else:
            return (
                f"Error: {self.error.error_name}\n"
                f"Description: {self.error.description}\n"
                f"Text: {self.error.text}\n"
                f"Details: {self.error.details} "
            )


def parse_error(error: Union[str, dict]) -> dict:
    if isinstance(error, dict) and "error" in error.keys():
        return error
    elif isinstance(error, str):
        return {"error": error}

    raise ValueError("Couldn't parse an error due to an invalid type")



import marshmallow_dataclass
from marshmallow import pre_load

from datagen.modalities.textual.identities.keypoints import base


@marshmallow_dataclass.dataclass(base_schema=base.KeypointsSchema)
class SceneKeypoints(base.SceneKeypoints):
    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs) -> dict:
        return {"scene": _rearange_nested_segments(in_data)}


def _rearange_nested_segments(in_data: dict) -> dict:
    converted_dict = {}
    for name, data in in_data.items():
        if isinstance(data, dict):
            if _is_single_coord_keypoint_segment(data):
                converted_dict[name] = _convert_to_single_coord(data)
            elif _is_keypoints_matrix_segment(data):
                converted_dict[name] = _convert_to_matrix(data)
            else:
                converted_dict[name] = _rearange_nested_segments(data)
        else:
            converted_dict[name] = data
    return converted_dict


def _is_single_coord_keypoint_segment(in_data: dict) -> bool:
    return all(key in in_data for key in ["pixel_2d", "global_3d", "is_visible"])


def _convert_to_single_coord(data):
    return {
        "coords_2d": data["pixel_2d"],
        "coords_3d": data["global_3d"],
        "is_visible": data["is_visible"] == "true",
    }


def _is_keypoints_matrix_segment(in_data: dict) -> bool:
    return all(s.isnumeric() for s in in_data.keys())


def _convert_to_matrix(kp_num_to_kp_coords: dict) -> dict:
    coords_2d_matrix, coords_3d_matrix, is_visible_arr = [], [], []
    kp_num_to_kp_coords = _convert_str_keys_to_int(kp_num_to_kp_coords)
    for _, kp_coords in sorted(kp_num_to_kp_coords.items()):
        coords_2d, coords_3d, is_visible_str = kp_coords["pixel_2d"], kp_coords["global_3d"], kp_coords["is_visible"]
        coords_2d_matrix.append([coords_2d["x"], coords_2d["y"]])
        coords_3d_matrix.append([coords_3d["x"], coords_3d["y"], coords_3d["z"]])
        is_visible_arr.append(is_visible_str == "true")
    return {"coords_2d": coords_2d_matrix, "coords_3d": coords_3d_matrix, "is_visible": is_visible_arr}


def _convert_str_keys_to_int(kp_num_to_kp_coords: dict) -> dict:
    return {int(kp_num): kp_coords for kp_num, kp_coords in kp_num_to_kp_coords.items()}



import numpy as np
from marshmallow.fields import Field
from marshmallow_dataclass import NewType


class _NumpyArray(Field):
    def _deserialize(self, value, *args, **kwargs):
        if isinstance(value, dict):
            if self._is_numeric_keys(value):
                value = self._extract_from_numeric_keys(value)
            else:
                value = self._extract_from_xyz_keys(value)
        return np.array(value)

    def _is_numeric_keys(self, value: dict) -> bool:
        return all(map(lambda key: key.isnumeric(), value.keys()))

    def _extract_from_numeric_keys(self, dict_: dict) -> list:
        list_ = []
        for idx in self._get_sorted_indexes(dict_):
            list_.append(dict_[str(idx)])
        return list_

    def _extract_list(self, dict_) -> list:
        try:
            return self._extract_from_xyz_keys(dict_)
        except KeyError:
            return self._extract_from_numeric_keys(dict_)

    def _extract_from_xyz_keys(self, dict_: dict) -> list:
        list_ = [dict_["x"], dict_["y"]]
        if "z" in dict_:
            list_.append(dict_["z"])
        return list_

    def _get_sorted_indexes(self, dict_) -> list:
        return sorted(int(s) for s in dict_.keys())


NumpyArray = NewType("NdArray", np.ndarray, field=_NumpyArray)



from dataclasses import dataclass
from pathlib import Path
from typing import List, Iterable

from dependency_injector import containers
from dependency_injector.wiring import inject

from datagen.components.datapoint import DataPoint

IDENTITIES_SCENE_FRAMES_RANGE = range(1)


@inject
@dataclass
class DatapointsRepository:
    datapoints_container: containers.DeclarativeContainer

    def get_datapoints(self, scene_path: Path, camera_name: str) -> List[DataPoint]:
        datapoints = []
        for frame_num in self._get_frames_range(scene_path):
            for environment in self._get_datapoints_environments(scene_path, camera_name, frame_num):
                datapoints.append(
                    self.datapoints_container.factory(
                        scene_path=scene_path,
                        camera=camera_name,
                        frame_num=frame_num,
                        visible_spectrum_image_name=environment.image_name,
                    )
                )
        return datapoints

    def _get_frames_range(self, scene_path: Path) -> range:
        if self._is_hic_scene(scene_path):
            return self._get_hic_scene_frames_range(scene_path)
        else:
            return IDENTITIES_SCENE_FRAMES_RANGE

    @staticmethod
    def _get_hic_scene_frames_range(scene_path: Path) -> range:
        frames_dirs = list(filter(lambda item: item.name.isnumeric(), scene_path.joinpath("frames").iterdir()))
        frames_num = len(frames_dirs)
        return range(1, frames_num + 1)

    def _get_datapoints_environments(
        self, scene_path: Path, camera_name: str, frame_num: int
    ) -> Iterable["Environment"]:
        return self.datapoints_container.modalities().read_textual_modality(
            modality_factory_name="environments",
            modality_file_path=self._get_environments_modality_file_path(scene_path, camera_name, frame_num),
        )

    def _get_environments_modality_file_path(self, scene_path: Path, camera_name: str, frame_num: int) -> str:
        if self._is_hic_scene(scene_path):
            base_path = scene_path.joinpath("frames", str(frame_num).zfill(3))
        else:
            base_path = scene_path
        return str(base_path.joinpath(camera_name, "environment.json"))

    @staticmethod
    def _is_hic_scene(scene_path: Path) -> bool:
        frames_path = scene_path.joinpath("frames")
        return frames_path.exists() and frames_path.is_dir()



from dataclasses import field
from typing import ClassVar, List

import marshmallow
import marshmallow_dataclass
from marshmallow.decorators import pre_load
from marshmallow.fields import Field
from marshmallow_dataclass import NewType

from datagen.modalities.textual.common.identity_label import IdentityLabel
from datagen.modalities.textual.common.ndarray import NumpyArray
from datagen.modalities.textual.common.sixdof import SixDOF

EYES_FULLY_OPENED = 5


@marshmallow_dataclass.dataclass
class FaceExpression:
    name: str
    intensity_level: int


@marshmallow_dataclass.dataclass
class RollPitchYaw:
    roll: float
    pitch: float
    yaw: float


@marshmallow_dataclass.dataclass
class HeadMetadata:
    head_root_location: NumpyArray
    head_rotation: RollPitchYaw
    head_six_dof: SixDOF


@marshmallow_dataclass.dataclass
class RollPitchYaw:
    roll: float
    pitch: float
    yaw: float


@marshmallow_dataclass.dataclass
class HeadMetadata:
    head_root_location: NumpyArray
    head_rotation: RollPitchYaw
    head_six_dof: SixDOF


@marshmallow_dataclass.dataclass
class Coord2DPerEye:
    camera_name: str
    right_eye: NumpyArray
    left_eye: NumpyArray


@marshmallow_dataclass.dataclass
class Coords3DPerEye:
    right_eye: NumpyArray
    left_eye: NumpyArray


class SinglePointPerEyeSchema(marshmallow.Schema):
    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs):
        in_data["coords_2d"] = self._adjust_to_2d_coords_schema(in_data.pop("2d"))
        in_data["coords_3d"] = in_data.pop("3d")
        return in_data

    @staticmethod
    def _adjust_to_2d_coords_schema(eyes_2d_data: dict) -> list:
        coord_2d_data = []
        for camera_name, camera_eyes_2d_data in eyes_2d_data.items():
            coord_2d_data.append({"camera_name": camera_name, **camera_eyes_2d_data})
        return coord_2d_data


@marshmallow_dataclass.dataclass(base_schema=SinglePointPerEyeSchema)
class SinglePointPerEye:
    coords_2d: List[Coord2DPerEye]
    coords_3d: Coords3DPerEye


@marshmallow_dataclass.dataclass
class Multi2DPointsPerEye:
    camera_name: str
    right_eye: NumpyArray
    left_eye: NumpyArray

    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs):
        in_data["right_eye"] = [[entry["x"], entry["y"]] for entry in in_data["right_eye"]]
        in_data["left_eye"] = [[entry["x"], entry["y"]] for entry in in_data["left_eye"]]
        return in_data


@marshmallow_dataclass.dataclass
class Multi3DPointsPerEye:
    right_eye: NumpyArray
    left_eye: NumpyArray

    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs):
        in_data["right_eye"] = [[entry["x"], entry["y"], entry["z"]] for entry in in_data["right_eye"]]
        in_data["left_eye"] = [[entry["x"], entry["y"], entry["z"]] for entry in in_data["left_eye"]]
        return in_data


class MultiPointsPerEyeSchema(marshmallow.Schema):
    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs):
        in_data["coords_2d"] = self._adjust_to_2d_coord_schema(in_data.pop("2d"))
        in_data["coords_3d"] = in_data.pop("3d")
        return in_data

    @staticmethod
    def _adjust_to_2d_coord_schema(eyes_2d_data: dict) -> list:
        coords_2d_data = []
        for camera_name, camera_eyes_2d_data in eyes_2d_data.items():
            coords_2d_data.append({"camera_name": camera_name, **camera_eyes_2d_data})
        return coords_2d_data


@marshmallow_dataclass.dataclass(base_schema=MultiPointsPerEyeSchema)
class MultiPointsPerEye:
    coords_2d: List[Multi2DPointsPerEye]
    coords_3d: Multi3DPointsPerEye


@marshmallow_dataclass.dataclass
class AxisDirection:
    optical_axis_direction: NumpyArray
    visual_axis_direction: NumpyArray

    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs):
        return in_data.pop("axis_directions")


@marshmallow_dataclass.dataclass
class AxisDirections:
    right_eye: AxisDirection
    left_eye: AxisDirection


@marshmallow_dataclass.dataclass
class EyeGaze:
    axis_directions: AxisDirections
    target_point: Coords3DPerEye
    eye_gaze_direction_type: str


@marshmallow_dataclass.dataclass
class Accessory:
    type: ClassVar[str]


@marshmallow_dataclass.dataclass
class Glasses(Accessory):
    type: ClassVar[str] = "glasses"
    style: str
    lens_color: str
    reflectivity: float
    transparency: float
    location: str = "on_nose"

    @pre_load
    def rearrange_fields(self, in_data: dict, **kwargs):
        if "metallic_intensity" in in_data:
            in_data["reflectivity"] = in_data.pop("metallic_intensity")
        if "transparency_intensity" in in_data:
            in_data["transparency"] = in_data.pop("transparency_intensity")
        return in_data

    @property
    def metallic_intensity(self) -> float:
        return self.reflectivity

    @property
    def transparency_intensity(self) -> float:
        return self.transparency


@marshmallow_dataclass.dataclass
class Mask(Accessory):
    type: ClassVar[str] = "mask"
    style: str
    mask_color: str
    mask_texture: str
    roughness: float
    location: str = "on_nose"


class AccessoryField(Field):
    ACCESSORY_TYPE_TO_CLASS = {"glasses": Glasses, "mask": Mask}

    def _deserialize(self, accessory_dict: dict, attr, data, **kwargs) -> Accessory:
        accessory_type = accessory_dict.pop("type")
        return self.ACCESSORY_TYPE_TO_CLASS[accessory_type.lower()].Schema().load(accessory_dict)


AccessoryType = NewType("Accessory", Accessory, field=AccessoryField)


@marshmallow_dataclass.dataclass
class ActorMetadata:
    identity_id: str
    identity_label: IdentityLabel
    facial_hair_included: bool
    face_expression: FaceExpression
    head_metadata: HeadMetadata
    center_of_rotation_point: SinglePointPerEye
    apex_of_cornea_point: SinglePointPerEye
    center_of_iris_point: SinglePointPerEye
    center_of_pupil_point: SinglePointPerEye
    iris_circle: MultiPointsPerEye
    pupil_circle: MultiPointsPerEye
    eye_gaze: EyeGaze
    eyelid_closure_intensity_level: int = field(default=EYES_FULLY_OPENED)
    accessories: List[AccessoryType] = field(default_factory=list)



from datagen.api import assets
from datagen.api.assets import Human
from datagen.api.catalog.impl import InitParametersHook


class HumansDefaultsHook(InitParametersHook[Human]):
    def __init__(self, asset_id_to_asset_defaults: dict):
        self._asset_id_to_asset_defaults = asset_id_to_asset_defaults

    def __call__(self, asset_id: str) -> dict:
        from datagen.api import catalog

        defaults = self._asset_id_to_asset_defaults[asset_id]
        default_eyes = catalog.eyes.parse(**defaults["eyes"])
        default_hair = catalog.hair.parse(**defaults["hair"])
        default_eyebrows = catalog.eyebrows.parse(**defaults["eyebrows"])
        return {"head": assets.Head(eyes=default_eyes, eyebrows=default_eyebrows, hair=default_hair)}



from pathlib import Path

from dynaconf import LazySettings

SETTINGS_TOML_PATH = Path(__file__).parent.joinpath("settings.toml")

settings = LazySettings(SETTINGS_FILE_FOR_DYNACONF=[SETTINGS_TOML_PATH])



import os

from dependency_injector import containers, providers

from datagen.api.client.session import ClientSession
from datagen.config import settings
from datagen.dev.logging import get_logger

logger = get_logger(__name__)

LONG_TIMEOUT_TIME = 36000  # 10 hours


def token_setup_instructions() -> str:
    if os.name == "nt":  # Windows
        return "set DG_AUTH_TOKEN=<your-auth-token>"
    else:  # Linux and MacOS
        return "export DG_AUTH_TOKEN=<your-auth-token>"


def get_token() -> str:
    token = os.environ.get("DG_AUTH_TOKEN")
    if not token:
        logger.warn(
            f"Authentication token is uninitialized. You can create a new token at "
            f"https://app.datagen.tech/token. \nAfter creating the token insert the following in your terminal:"
            f" {token_setup_instructions()}"
        )

    return token


class SessionContainer(containers.DeclarativeContainer):

    client_session = providers.Factory(
        ClientSession, base_url=settings["url"]["base"], headers={"authorization_token": get_token()}
    )

    long_timeout = providers.Object(LONG_TIMEOUT_TIME)
